// Copyright (c) 1997 Andrew Maltsev <am@amsoft.ru>
// All rights reserved
// 
// Definitions of internal methods of EasySQL library.
// See License for your rights and obligations. Or mail to author.
//
// $Id: es_int.h,v 1.10 1997/10/23 18:40:30 am Exp $
//
#ifndef __AM_es_int_h
#define __AM_es_int_h
#include <stdio.h>
#include <stdarg.h>
#include "easysql.h"

#ifdef __GNUC__
#define	attribute(x) __attribute__((x))
#else
#define	attribute(x) /* x */
#endif

///////////////////////////////////////////////////////////////////////////////
// Driver interface protocol version
//
#define EasySQL_ProtocolVersion	1

///////////////////////////////////////////////////////////////////////////////
// Default retry delay
//
#define DefaultRetryDelay	60

///////////////////////////////////////////////////////////////////////////////
// Dynamic loader (realization of methods IS os-dependent, see src/dynaload.C!)
//
class easysql_dynamic_loader : public easysql_error_reporting
{ void *handle;
 public:
  easysql_dynamic_loader(char const *file=NULL);
  ~easysql_dynamic_loader();
  operator bool();
  bool load(char const *file);
  bool unload();
  void *symbol(char const *sym);
};

///////////////////////////////////////////////////////////////////////////////
// Additional arguments to be given to SQL drivers. A way to set up
// port, socket, open mode, etc..
//
class easysql_arguments
{ char **values;
  char **names;
  unsigned num;
 public:
  easysql_arguments(char const *args=0);	// Accepts a string in form of
  						// "arg1=value1 arg2=value2 ..."
  ~easysql_arguments();
  void reset();					// Cleaning
  bool merge(char const *args);			// Same as in constructor
  bool merge(char const *name,			// Name and value separately
             char const *value);
  static
   char const *value(easysql_arguments const *,	// Interface method
                     char const *name);
};

///////////////////////////////////////////////////////////////////////////////
// Driver loader. Base for childrens, who are responsible for getting
// driver from on disk object or from in core register.
//
class easysql_driver_loader
{protected:
  void *(*driver)(char const *);
 public:
  easysql_driver_loader();
  virtual ~easysql_driver_loader();
  void * operator() (char const *par) const;
  operator bool() const;
};

///////////////////////////////////////////////////////////////////////////////
// Shell for the driver. Builds suitable driver object by getting
// interface function from on disk shared object or from in-core
// register; doing some tests and asking for real functions.
//
class easysql_driver_shell : public easysql_error_reporting
{ easysql_driver_loader &drv;
  void *data;					// Filled from driver
  char const *(*i_version)(void *);
  void *(*i_create)(void);
  bool (*i_finish)(void *);
  bool (*i_connect)(void *, char const *(*)(easysql_arguments const *, char const *), void const *);
  bool (*i_query)(void *, esqlResult &r, char const *q);
  bool (*i_rfree)(void *, esqlResult r);
  bool (*i_row)(void *, esqlRow &r, esqlResult res);
  bool (*i_drop)(void *);
  char const *(*i_error)(void *);
  void (*i_no_error)(void *);
  char const *(*i_status)(void *);
  bool (*i_check_connection)(void *);
  char const *i_to_be_escaped;
 public:
  easysql_driver_shell(char const *name);
  ~easysql_driver_shell();
  char const *error_text();		// Overriding easysql_error_reporting
  void no_error();
  operator bool();
  char const *version()
   { no_error();
     return i_version(data);
   }
  bool connect(easysql_arguments const &args)
   { no_error();
     return i_connect(data,easysql_arguments::value,&args);
   }
  bool query(esqlResult &r, char const *q)
   { no_error();
     return i_query(data,r,q);
   }
  bool rfree(esqlResult r)
   { no_error();
     return i_rfree(data,r);
   }
  bool row(esqlRow &r, esqlResult res)
   { no_error();
     return i_row(data,r,res);
   }
  bool drop()
   { no_error();
     return i_drop(data);
   }
  char const *status()
   { no_error();
     return i_status(data);
   }
  bool check_connection()
   { no_error();
     return i_check_connection(data);
   }
  char const *escaped_chars()
   { no_error();
     return i_to_be_escaped;
   }
};

///////////////////////////////////////////////////////////////////////////////
// Queries register. Not really required, but implemented only as a
// reminder and automated query freeing tool. Frees all not rfree'd queries
// on destroing easysql
//
class easysql_queries_register
{ struct
   { esqlResult r;
     bool active;
   } *list;
  unsigned size;
  easysql_driver_shell *driver;
 public:
  easysql_queries_register(easysql_driver_shell *drv);
  ~easysql_queries_register();
  void add(esqlResult r);
  bool drop(esqlResult r);
};

///////////////////////////////////////////////////////////////////////////////
// Driver's register. Will be automatically filled when used on top level in
// driver's code
//
class easysql_register
{ struct driverdata
   { char const *name;
     void *(*iface)(char const *);
   };
  static driverdata *data;
  static unsigned ndata;
 public:
  easysql_register(char const *name,
                   void *(*)(char const *));	// Registration from drivers
  ~easysql_register();
  static void *(*find(char const *name)		// Looking up from loaders
        )(char const *);
};

///////////////////////////////////////////////////////////////////////////////
// Driver's declaration to be used in driver's source code. Will
// look like this:
//  DeclareDriver(drvname);
//
#define DeclareDriver(drvname,iface) \
		easysql_register drvname##_register_entry(#drvname,iface)

///////////////////////////////////////////////////////////////////////////////
// Debugging..
//
#define D_L1		EASYSQL_D_L1
#define D_L2		EASYSQL_D_L2
#define	D_L3		EASYSQL_D_L3
#define D_L4		EASYSQL_D_L4
//
#define D_MISC		EASYSQL_D_MISC
#define D_SQL		EASYSQL_D_SQL
#define D_DRV		EASYSQL_D_DRV
#define D_IDRV		EASYSQL_D_IDRV
#define D_C		EASYSQL_D_C
//
#define D_ALL		EASYSQL_D_ALL
#define D_NONE		EASYSQL_D_NONE
#define D_ERROR		EASYSQL_D_ERROR
//
#define	log		easysql_log
//
void easysql_log(unsigned level, char const *fmt, ...) attribute(format(printf,2,3));

/////////////////////////// SUPPORT FUNCTIONS /////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Configuration parameters
//
// Constants
//
char const * const defaultHomeDir	="/usr/local/easysql";
char const * const configFile		="/usr/local/etc/easysql.conf";
char const * const configFile2		="/etc/easysql.conf";
//
// Some precautions to exclude overlapping with user variables but allow easy
// programming of API itself..
//
#define homeDir			_easysql_homeDir
#define driversDir		_easysql_driversDir
#define dbDir			_easysql_dbDir
#define allowUserConfigs	_easysql_allowUCf
#define getConfig		_easysql_getConfig
//
// References
//
extern char const *homeDir;
extern char const *driversDir;
extern char const *dbDir;
extern bool allowUserConfigs;
//
// Reading of configuration
//
bool getConfig();

///////////////////////////////////////////////////////////////////////////////
// Unlimited formatted string in dynamically allocated buffer
//
#define vaprintf	_easysql_vaprintf
#define qprintf		_easysql_qprintf
char *vaprintf(char const *fmt, va_list ap) attribute(format(printf,1,0));
char const *qprintf(char const *fmt, ...) attribute(format(printf,1,2));

///////////////////////////////////////////////////////////////////////////////
// Memory operations. Safe and will only return legal result or no
// result at all. So there is no need to check for NULL after mAlloc -
// it's generally suitable on Unix-like oses.
//
#define	Free		_easysql_free
#define	mAlloc		_easysql_malloc
#define	reAlloc		_easysql_realloc
#define strDup		_easysql_strdup
void Free(void *);
void *mAlloc(unsigned long);
void *reAlloc(void *, unsigned long);
char *strDup(char const *);

//////////////////////////////////////////////////////////////////////////////
// Helper to define rcsid's
//
#define R(x) static char am_rcsid[]=x; \
             static void *const am_use_rcsid=&am_rcsid
extern bool rcsid;

#undef attribute
#endif	// __AM_es_int_h
